"""Module to evaluate blocking when ground truth is available."""
from tqdm import tqdm


def assess_blocks_2party(filtered_reverse_indices, data):
    """Assess pair completeness and reduction ratio of blocking result.

    :ivar filtered_reverse_indices for each data provider, a dict containing the mapping from block id to corresponding record ids.
    :ivar data: a list of lists of entity_ids for 2 data providers
    """
    # currently just support for two party
    dp1_signature, dp2_signature = filtered_reverse_indices
    dp1_data, dp2_data = data

    cand_pairs = {}  # key is record id of data provider1 and value is list of record ids from data provider 2
    num_block_true_matches = 0
    num_block_false_matches = 0

    keys = set(dp1_signature.keys()).intersection(dp2_signature.keys())
    for key in tqdm(keys, desc='assessing blocks', total=len(keys), unit='key'):
        dp1_recs = dp1_signature.get(key, None)
        dp2_recs = dp2_signature.get(key, None)
        if dp1_recs is None or dp2_recs is None:
            continue
        for d1 in dp1_recs:
            d1_entity = dp1_data[d1]
            d1_cache = cand_pairs.get(d1_entity, set())
            for d2 in dp2_recs:
                d2_entity = dp2_data[d2]
                if d2_entity not in d1_cache:
                    d1_cache.add(d2_entity)
                    if d2_entity == d1_entity:
                        num_block_true_matches += 1
                    else:
                        num_block_false_matches += 1
            cand_pairs[d1_entity] = d1_cache

    num_cand_rec_pairs = num_block_true_matches + num_block_false_matches
    total_rec = len(dp1_data) * len(dp2_data)

    entity1 = set(dp1_data)
    entity2 = set(dp2_data)
    num_all_true_matches = len(entity1.intersection(entity2))

    # pair completeness is the "recall" before matching stage
    rr = 1.0 - float(num_cand_rec_pairs) / total_rec
    pc = float(num_block_true_matches) / num_all_true_matches
    return rr, pc



import base64
from bitarray import bitarray
from typing import Sequence, Any, List


def check_header(header: List[str], row: Sequence[Any]):
    """Check if header is consistent with data dimension."""
    error_msg = f'There are {len(header)} features specified in header, but each row has {len(row)} fields'
    assert len(header) == len(row), error_msg


def deserialize_bitarray(bytes_data: Any):
    ba = bitarray(endian='big')
    data_as_bytes = base64.decodebytes(bytes_data.encode())
    ba.frombytes(data_as_bytes)
    return ba


def deserialize_filters(filters: Sequence[Any]):
    res = []
    for i, f in enumerate(filters):
        ba = deserialize_bitarray(f)
        res.append(ba)
    return res




import random
from typing import Any, Dict, List, Sequence, Optional
from blocklib.configuration import get_config
from blocklib.stats import reversed_index_stats


class PPRLIndex:
    """Base class for PPRL indexing/blocking."""

    def __init__(self, config: Dict = {}) -> None:
        """Initialise base class."""
        self.rec_dict = None
        self.ent_id_col = None
        self.rec_id_col = None
        self.stats = {}  # type: Dict[str, Any]

    def build_reversed_index(self, data: Sequence[Sequence], verbose: bool, header: Optional[List[str]]  = None):
        """Method which builds the index for all database.

           :param data: list of tuples, PII dataset
           :param verbose: print additional information to std out.
           :param header: file header, optional

           See derived classes for actual implementations.
        """
        raise NotImplementedError("Derived class needs to implement")

    def summarize_reversed_index(self, reversed_index: Dict):
        """Summarize statistics of reverted index / blocks."""
        assert len(reversed_index) > 0
        # statistics of block
        self.stats = reversed_index_stats(reversed_index)
        # find how many blocks each entity / record is a member of
        rec_to_block = {}  # type: Dict[Any, List[Any]]
        for block_id, block in reversed_index.items():
            for rec in block:
                if rec in rec_to_block:
                    rec_to_block[rec].append(block_id)
                else:
                    rec_to_block[rec] = [block_id]
        num_of_blocks_per_rec = [len(x) for x in rec_to_block.values()]
        self.stats['num_of_blocks_per_rec'] = num_of_blocks_per_rec

        print('Statistics for the generated blocks:')
        print('\tNumber of Blocks:   {}'.format(self.stats['num_of_blocks']))
        print('\tMinimum Block Size: {}'.format(self.stats['min_size']))
        print('\tMaximum Block Size: {}'.format(self.stats['max_size']))
        print('\tAverage Block Size: {}'.format(self.stats['avg_size']))
        print('\tMedian Block Size:  {}'.format(self.stats['med_size']))
        print('\tStandard Deviation of Block Size:  {}'.format(self.stats['std_size']))

        return self.stats

    def select_reference_value(self, reference_data: Sequence[Sequence], ref_data_config: Dict):
        """Load reference data for methods need reference."""
        # read configurations
        ref_default_features = get_config(ref_data_config, 'blocking-features')
        ref_random_seed = get_config(ref_data_config, 'random-state')
        num_vals = get_config(ref_data_config, 'num-reference-values')

        # extract features in config
        rec_features = [''.join([dtuple[x] for x in ref_default_features]) for dtuple in reference_data]

        # generate reference values
        random.seed(ref_random_seed)
        ref_val_list = random.sample(rec_features, num_vals)

        print('  Selected %d random reference values' % (len(ref_val_list)))
        return ref_val_list



import statistics
from collections import defaultdict
from typing import Dict, List, Sequence, Any, Optional

import numpy as np

from .configuration import get_config
from .encoding import flip_bloom_filter
from .pprlindex import PPRLIndex
from .signature_generator import generate_signatures
from .stats import reversed_index_per_strategy_stats
from .utils import check_header


class PPRLIndexPSignature(PPRLIndex):
    """Class that implements the PPRL indexing technique:

        Reference scalability entity resolution using probability signatures
        on parallel databases.

        This class includes an implementation of p-sig algorithm.
    """

    def __init__(self, config: Dict) -> None:
        """Initialize the class and set the required parameters.

        Arguments:
        - config: dict
            Configuration for P-Sig reverted index.

        """
        super().__init__()
        self.filter_config = get_config(config, "filter")
        self.blocking_config = get_config(config, "blocking-filter")
        self.signature_strategies = get_config(config, 'signatureSpecs')
        self.rec_id_col = config.get("record-id-col", None)

    def build_reversed_index(self, data: Sequence[Sequence], verbose: bool = False, header: Optional[List[str]] = None):
        """Build inverted index given P-Sig method."""
        # find blocking feature index if blocking feature type is string
        feature_to_index = None
        if header:
            check_header(header, data[0])
            feature_to_index = {name: ind for ind, name in enumerate(header)}

        # Build index of records
        if self.rec_id_col is None:
            record_ids = np.arange(len(data))
        else:
            record_ids = [x[self.rec_id_col] for x in data]

        reversed_index_per_strategy = \
            [defaultdict(list) for _ in range(len(self.signature_strategies))]  # type: List[Dict[str, List[Any]]]
        # Build inverted index
        # {signature -> record ids}
        for rec_id, dtuple in zip(record_ids, data):

            signatures = generate_signatures(self.signature_strategies, dtuple, feature_to_index)

            for i, signature in enumerate(signatures):
                reversed_index_per_strategy[i][signature].append(rec_id)

        n = len(data)
        reversed_index_per_strategy = [self.filter_reversed_index(data, reversed_index) for reversed_index in
                                       reversed_index_per_strategy]
        if verbose:
            strat_stats = reversed_index_per_strategy_stats(reversed_index_per_strategy, n)
            print("Statistics for the individual strategies:")
            for strat_stat in strat_stats:
                print('Strategy {}:'.format(strat_stat["strategy_idx"]))
                print('\tblock size {} min, {} max, {:.2f} avg, {} median, {:.2f} std'
                      .format(strat_stat["min_size"], strat_stat["max_size"], strat_stat["avg_size"],
                              strat_stat["med_size"], strat_stat["std_size"]))
                print('\t {} blocks, {} filtered elements, {:.2f}% coverage'
                      .format(strat_stat["num_of_blocks"], strat_stat["num_filtered_elements"],
                              (strat_stat["coverage"] * 100)))

        # combine the reversed indices into one
        filtered_reversed_index = reversed_index_per_strategy[0]
        for rev_idx in reversed_index_per_strategy[1:]:
            filtered_reversed_index.update(rev_idx)

        # check if final inverted index is empty
        if len(filtered_reversed_index) == 0:
            raise ValueError('P-Sig: All records are filtered out!')

        # compute coverage information
        entities = set()
        for recids in filtered_reversed_index.values():
            for rid in recids:
                entities.add(rid)
        coverage = round(len(entities) / len(record_ids) * 100, 2)
        if coverage == 100:
            print('P-Sig: {}% records are covered in blocks'.format(coverage))
        else:
            print('P-Sig: Warning! only {}% records are covered in blocks. Please consider to improve signatures'
                  .format(coverage))

        # map signatures in reversed_index into bloom filter
        num_hash_func = int(self.blocking_config.get("number-hash-functions", None))
        bf_len = int(self.blocking_config.get("bf-len", None))

        reversed_index = {}  # type: Dict[str, List[Any]]

        for signature, rec_ids in filtered_reversed_index.items():
            bf_set = str(tuple(flip_bloom_filter(signature, bf_len, num_hash_func)))
            if bf_set in reversed_index:
                reversed_index[bf_set].extend(rec_ids)
            else:
                reversed_index[bf_set] = rec_ids

        return reversed_index

    def filter_reversed_index(self, data: Sequence[Sequence], reversed_index: Dict):
        # Filter inverted index based on ratio
        n = len(data)

        # filter blocks based on filter type
        filter_type = get_config(self.filter_config, "type")
        if filter_type == "ratio":
            min_occur_ratio = get_config(self.filter_config, 'min')
            max_occur_ratio = get_config(self.filter_config, 'max')
            reversed_index = {k: v for k, v in reversed_index.items() if n * max_occur_ratio > len(v) > n * min_occur_ratio}
        elif filter_type == "count":
            min_occur_count = get_config(self.filter_config, "min")
            max_occur_count = get_config(self.filter_config, "max")
            reversed_index = {k: v for k, v in reversed_index.items() if max_occur_count > len(v) > min_occur_count}
        else:
            raise NotImplementedError("Don't support {} filter yet.".format(filter_type))

        return reversed_index



import json
from typing import Dict
import jsonschema
import pathlib


def load_schema(file_name: str):
    path = pathlib.Path(__file__).parent / 'schemas' / file_name

    # schema_bytes = pkgutil.get_data('anonlinkclient', 'docs/schemas/{}'.format(file_name))
    with open(str(path.resolve()), 'rt') as schema_file:
        try:
            return json.load(schema_file)
        except json.decoder.JSONDecodeError as e:
            raise ValueError("Invalid schema") from e



def validate_signature_config(config: Dict):
    schema = load_schema('signature-config-schema.json')
    try:
        jsonschema.validate(config, schema)
    except jsonschema.exceptions.ValidationError as e:
        raise ValueError('The signature config is not valid.\n\n' + str(e)) from e



import random
import numpy as np
from collections import defaultdict
from typing import Dict, Sequence, Any, List, Optional
from blocklib.configuration import get_config
from .pprlindex import PPRLIndex
from .encoding import generate_bloom_filter
from .utils import deserialize_filters, check_header


class PPRLIndexLambdaFold(PPRLIndex):
    """Class that implements the PPRL indexing technique:

        An LSH-Based Blocking Approach with a Homomorphic Matching Technique for Privacy-Preserving Record Linkage.

        This class includes an implementation of Lambda-fold redundant blocking method.
    """

    def __init__(self, config: Dict):
        """Initialize the class and set the required parameters.

        Arguments:
        - config: dict
            Configuration for P-Sig reverted index.

        """
        super().__init__()
        # Lambda: number of redundant tables
        self.mylambda = int(get_config(config, "Lambda"))
        # bf-len: length of bloom filter
        self.bf_len = int(get_config(config, "bf-len"))
        # num_hash_function
        self.num_hash_function = int(get_config(config, "num-hash-funcs"))
        # K: number of base Hamming LSH hashing functions
        self.K = int(get_config(config, "K"))
        # blocking-features: list of blocking feature indices
        self.blocking_features = get_config(config, "blocking-features")
        self.input_clks = get_config(config, 'input-clks')
        self.random_state = get_config(config, "random_state")
        self.record_id_col = config.get("record-id-col", None)

    def __record_to_bf__(self, record: Sequence):
        """Convert a record to list of bigrams and then map to a bloom filter."""
        s = ''.join([record[i] for i in self.blocking_features])
        # generate list of bigram of s. hash each bigram to position of bit 1 and flip bloom filter
        ngram = 2
        grams = [s[i: i + ngram] for i in range(len(s) - ngram + 1)]
        bloom_filter = generate_bloom_filter(grams, self.bf_len, self.num_hash_function)
        return bloom_filter

    def build_reversed_index(self, data: Sequence[Any], verbose: bool = False, header: Optional[List[str]] = None):
        """Build inverted index for PPRL Lambda-fold blocking method.

        :param data: list of lists
        :param verbose: ignored
        :return:
        """
        # convert blocking feature to index if header is given
        if header:
            check_header(header, data[0])
            feature_to_index = {name: ind for ind, name in enumerate(header)}
            self.blocking_features = [feature_to_index[x] for x in self.blocking_features]

        # create record index lists
        if self.record_id_col is None:
            record_ids = list(range(len(data)))
        else:
            record_ids = [x[self.record_id_col] for x in data]

        random.seed(self.random_state)

        if self.input_clks:
            clks = deserialize_filters(data)
        else:
            clks = [self.__record_to_bf__(rec) for rec in data]
        bf_len = len(clks[0])

        # build Lambda fold tables and add to the invert index
        invert_index = {}  # type: Dict[Any, List[Any]]
        for i in range(self.mylambda):
            lambda_table = defaultdict(list)  # type: Dict[Any, Any]
            # sample K indices from [0, bf-len]
            indices = random.sample(range(bf_len), self.K)
            for rec_id, clk in zip(record_ids, clks):
                block_key = ''.join(['1' if clk[ind] else '0' for ind in indices])
                lambda_table['{}{}'.format(i, block_key)].append(rec_id)
            invert_index.update(lambda_table)

        return invert_index




import statistics
from typing import Sequence, Dict, List, Any


def reversed_index_per_strategy_stats(reversed_index_per_strategy: Sequence[Dict[str, List[Any]]], num_elements: int):
    strat_stats = []
    for i, reversed_index in enumerate(reversed_index_per_strategy):
        stats = reversed_index_stats(reversed_index)
        stats['strategy_idx'] = i
        _add_coverage_to_stats_per_stragegy(stats, num_elements)
        strat_stats.append(stats)
    return strat_stats


def _add_coverage_to_stats_per_stragegy(stats: Dict, num_elements: int):
    # shortcut, as each element can only be part of one block, we can just use the sum of the blocks
    stats['num_filtered_elements'] = num_elements - stats['sum_of_blocks']
    stats['coverage'] = 0 if stats['sum_of_blocks'] == 0 else stats['sum_of_blocks'] / num_elements


def reversed_index_stats(reversed_index: Dict[str, List[Any]]):
    lengths = [len(rv) for rv in reversed_index.values()]
    stats = {
        'num_of_blocks': len(lengths),
        'min_size': 0 if len(lengths) == 0 else min(lengths),
        'max_size': 0 if len(lengths) == 0 else max(lengths),
        'avg_size': 0 if len(lengths) == 0 else statistics.mean(lengths),
        'med_size': 0 if len(lengths) == 0 else int(statistics.median(lengths)),
        'std_size': 0 if 0 <= len(lengths) <= 1 else statistics.stdev(lengths),
        'sum_of_blocks': sum(lengths)
    }
    return stats



from typing import Any, Callable, Dict, List, Sequence, Optional

from metaphone import doublemetaphone


def generate_by_feature_value(attr_ind: int, dtuple: Sequence):
    """Generate signatures by simply return original feature at attr_ind."""
    return dtuple[attr_ind]


def generate_by_char_at(attr_ind: int, dtuple: Sequence, pos: List[Any]):
    """ Generate signatures by select subset of characters in original features.
    >>> res = generate_by_char_at(2, ('harry potter', '4 Privet Drive', 'Little Whinging', 'Surrey'), [0, 3])
    >>> assert res == 'Lt'
    >>> res = generate_by_char_at(2, ('harry potter', '4 Privet Drive', 'Little Whinging', 'Surrey'), [":4"])
    >>> assert res == 'Litt'
    """
    sig = []
    feature = dtuple[attr_ind]

    # missing value
    if feature == '':
        return None

    max_ind = len(feature)
    for p in pos:
        if type(p) == int:
            p = min(p, max_ind - 1)
            sig.append(feature[p])
        elif ':' not in p:
            p = int(p)
            p = min(p, max_ind - 1)
            sig.append(feature[p])
        else:
            start_ind, end_ind = p.split(":")
            if start_ind != '' and end_ind != '':
                start_ind = int(start_ind)
                end_ind = int(end_ind)
                assert start_ind < end_ind, "Start index should be less than End index in {}".format(p)
                start_ind = min(start_ind, max_ind - 1)
                end_ind = min(end_ind, max_ind)
                c = feature[start_ind: end_ind]
            elif start_ind == '' and end_ind != '':
                end_ind = int(end_ind)
                end_ind = min(end_ind, max_ind)
                c = feature[:end_ind]
            elif start_ind != '' and end_ind == '':
                start_ind = int(start_ind)
                start_ind = min(start_ind, max_ind)
                c = feature[start_ind:]
            else:
                raise ValueError('Invalid pos argument: {}'.format(p))
            sig.append(c)

    return ''.join(sig)


def generate_by_metaphone(attr_ind: int, dtuple: Sequence):
    """Generate a phonetic encoding of features using metaphone.

    >>> generate_by_metaphone(0, ('Smith', 'Schmidt', 2134))
    'SM0XMT'

    """
    feature = dtuple[attr_ind]
    phonetic_encoding = doublemetaphone(feature)
    return ''.join(phonetic_encoding)


#################################################
########## Add strategy here ####################
#################################################
SIGNATURE_STRATEGIES = {
    'feature-value': generate_by_feature_value,
    "characters-at": generate_by_char_at,
    "characters_at": generate_by_char_at,
    'metaphone': generate_by_metaphone,
}  # type: Dict[str, Callable[..., str]]


def generate_signatures(signature_strategies: List[List],
                        dtuple: Sequence,
                        feature_to_index: Optional[Dict[str, int]] = None):
    """Generate signatures for one record.

    :param signature_strategies:
        A list of dicts each describing a strategy to generate signatures.

    :param dtuple:
        Raw data to generate signatures from

    :param feature_to_index:
        Mapping from feature name to feature index

    :return signatures: set of str
    """
    # signatures to return
    signatures = []

    # loop through each strategy

    for i, strategy in enumerate(signature_strategies):
        sig = []
        for spec in strategy:
            # arguments that we need to pass for any strategy
            attr = spec.get("feature", -1)
            if type(attr) == str:
                assert feature_to_index
                attr_ind = feature_to_index.get(attr, None)
                if attr_ind is None:
                    raise ValueError(f'Feature {attr} is not in the dataset')
            else:
                attr_ind = attr
            args = dict(attr_ind=attr_ind, dtuple=[str(x) for x in dtuple])
            config = spec.get('config', {})

            # find the correct strategy function to call
            func = SIGNATURE_STRATEGIES.get(spec['type'], None)

            if func is None:
                strategy_type = spec['type']
                print(spec)
                raise NotImplementedError('Strategy {} is not implemented yet!'.format(strategy_type))
            else:
                config.update(args)
                s = func(**config)
                sig.append(s)
        signatures.append('{}_{}'.format(i, "_".join([x for x in sig if x is not None])))

    return signatures



"""Class to implement privacy preserving encoding."""
import hashlib
import numpy as np
from typing import List, Set


def flip_bloom_filter(string: str, bf_len: int, num_hash_funct: int):
    """
    Hash string and return indices of bits that have been flipped correspondingly.

    :param string: string: to be hashed and to flip bloom filter
    :param bf_len: int: length of bloom filter
    :param num_hash_funct: int: number of hash functions
    :return: bfset: a set of integers - indices that have been flipped to 1
    """
    # config for hashing
    h1 = hashlib.sha1
    h2 = hashlib.md5

    sha_bytes = h1(string.encode('utf-8')).digest()
    md5_bytes = h2(string.encode('utf-8')).digest()
    int1 = int.from_bytes(sha_bytes, 'big') % bf_len
    int2 = int.from_bytes(md5_bytes, 'big') % bf_len

    # flip {num_hash_funct} times
    bfset = set()
    for i in range(num_hash_funct):
        gi = (int1 + i * int2) % bf_len
        bfset.add(gi)

    return bfset


def generate_bloom_filter(list_of_strs: List[str], bf_len: int, num_hash_funct: int):
    """
    Generate a bloom filter given list of strings.

    :param return_cbf_index_sig_map:
    :param list_of_strs:
    :param bf_len:
    :param num_hash_funct:
    :return: bloom_filter_vector if return_cbf_index_sig_map is False else (bloom_filter_vector, cbf_index_sig_map)
    """
    # go through each signature and generate bloom filter of it
    # -- we only store the set of index that flipped to 1
    candidate_bloom_filter = set()  # type: Set[int]

    for signature in list_of_strs:
        bfset = flip_bloom_filter(signature, bf_len, num_hash_funct)
        # union indices that have been flipped 1 in candidate bf
        candidate_bloom_filter = candidate_bloom_filter.union(bfset)

    # massage the cbf into a numpy bool array from a set
    bloom_filter_vector = np.zeros(bf_len, dtype=bool)
    bloom_filter_vector[list(candidate_bloom_filter)] = True

    return bloom_filter_vector



"""Similarity Measure Algorithms."""
from blocklib.configuration import get_config
import logging
from abc import ABC
from typing import Dict, List, Tuple


class SimMeasure(ABC):
    """General class that implements similarity measures.
    """

    def sim(self, s1: str, s2: str, cache: bool = False):
        """Calculate and return the similarity between two strings (a value
           between 0 and 1).

           If this similarity should be cached set the argument cache to True.
        """


class EditSim(SimMeasure):
    """Class that implements Edit (or Levenshtein) distance for two strings."""

    def __init__(self, config: Dict):
        """Initialize instance."""
        super().__init__()
        if 'min_threshold' in config:
            self.min_threshold = config['min_threshold']
        else:
            self.min_threshold = None

    def sim(self, str1: str, str2: str, cache: bool = False):
        """Return sim score between 0 to 1.
        """
        min_threshold = self.min_threshold
        # Quick check if the strings are empty or the same
        if (str1 == '') or (str2 == ''):
            return 0.0
        elif str1 == str2:
            return 1.0

        n = len(str1)
        m = len(str2)
        max_len = max(n, m)

        if min_threshold is not None:
            if isinstance(min_threshold, float) and 0 <= min_threshold <= 1:

                len_diff = abs(n - m)
                w = 1.0 - float(len_diff) / float(max_len)

                if w < min_threshold:
                    return 0.0  # Similarity is smaller than minimum threshold

                else:  # Calculate the maximum distance possible with this threshold
                    max_dist = (1.0 - min_threshold) * max_len

            else:
                msg = 'Illegal value for minimum threshold (not between 0 and 1): {}'.format(min_threshold)
                raise ValueError(msg)

        if n > m:  # Make sure n <= m, to use O(min(n,m)) space
            str1, str2 = str2, str1
            n, m = m, n

        current = list(range(n + 1))

        for i in range(1, m + 1):
            previous = current
            current = [i] + n * [0]
            str2char = str2[i - 1]

            for j in range(1, n + 1):
                substitute = previous[j - 1]
                if str1[j - 1] != str2char:
                    substitute += 1

                # Get minimum of insert, delete and substitute
                current[j] = min(previous[j] + 1, current[j - 1] + 1, substitute)

            if (min_threshold is not None) and (min(current) > max_dist):
                return 1.0 - float(max_dist + 1) / float(max_len)

        w = 1.0 - float(current[n]) / float(max_len)
        return w


class DiceSim(SimMeasure):
    """Class that implements the Dice coefficient for the two input strings.
       This methods uses the constants: ngram_len and ngram_padding (and if this
       constant is set to True also padding_start_char and self.padding_end_char).

       If the argument cache is set to True then the generated q-gram lists
       will be stored in a dictionary to prevent their repeated computation.
    """

    def __init__(self, config: Dict):
        """Initialise the cache."""
        self.ngram_len = int(get_config(config, 'ngram_len'))
        self.ngram_padding = get_config(config, 'ngram_padding')
        self.padding_start_char = get_config(config, 'padding_start_char')
        self.padding_end_char = get_config(config, 'padding_end_char')

        # Store strings converted into q-grams. Keys in this will be strings and their values their q-gram list
        self.q_gram_cache = {}  # type: Dict[str, List[str]]

        # Store the string pair and its similarity in a cache as well
        self.sim_cache = {}  # type: Dict[Tuple[str, str], float]

    def sim(self, s1: str, s2: str, cache: bool = False):
        """Calculate the similarity between the given two strings. The method
           returns a value between 0.0 and 1.0.

           If this similarity should be cached set the argument cache to True.
        """

        if s1 == s2:  # Quick check for equality
            return 1.0

        # Check if the string pair has been compared before
        if (s1, s2) in self.sim_cache:
            return self.sim_cache[(s1, s2)]

        q_minus_1 = self.ngram_len - 1

        # Convert input strings into q-gram lists
        l1 = self._convert_to_qgrams(s1, q_minus_1, cache)
        l2 = self._convert_to_qgrams(s2, q_minus_1, cache)

        common = len(set(l1).intersection(set(l2)))

        sim = 2.0 * common / (len(l1) + len(l2))

        if cache:
            self.sim_cache[(s1, s2)] = sim

        return sim

    def _convert_to_qgrams(self, inputstr: str, q_minus_1: int, cache: bool):
        if cache and (inputstr in self.q_gram_cache):
            qgrams = self.q_gram_cache[inputstr]

        else:
            # Need to calculate q-gram list for the first string
            if self.ngram_padding:
                ps1 = self.padding_start_char * q_minus_1 + inputstr + self.padding_end_char * q_minus_1
            else:
                ps1 = inputstr

            qgrams = [ps1[i:i + self.ngram_len] for i in range(len(ps1) - q_minus_1)]

            if cache:
                self.q_gram_cache[inputstr] = qgrams
        return qgrams




import setuptools

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

requirements = [
    "jsonschema>=3.1",
    "numpy>=1.17",
    "metaphone>=0.6",
    "tqdm>=4.36.1",
    "bitarray-hardbyte>=1.2"
]

setuptools.setup(
    name="blocklib",
    version="0.1.5",
    author="Joyce Wang",
    author_email="joyce.wang@csiro.au",
    description="A library for blocking in record linkage",
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/data61/blocklib',
    license='Apache',
    packages=setuptools.find_packages(),
    package_data={'blocklib': ['schemas/*.json*']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    tests_require=[
        "pytest>=5.0",
    ]
)



from typing import Dict


def get_config(config: Dict, arg_name: str):
    """Get arg value if arg_name exists in the config.

    Arguments
    ---------
    config: dict
    arg_name: str

    """
    value = config.get(arg_name, 'not specified')
    if value == 'not specified':
        raise ValueError('Argument "{}" was not specified\n\n{}'.format(arg_name, config))
    return value



"""Class that implement candidate block generations."""
from typing import Dict, Sequence, Tuple, Type, List, Optional
from .configuration import get_config
from .pprlindex import PPRLIndex
from .pprlpsig import PPRLIndexPSignature
from .pprllambdafold import PPRLIndexLambdaFold
from .validation import validate_signature_config


PPRLSTATES = {
    "p-sig": PPRLIndexPSignature,
    "lambda-fold": PPRLIndexLambdaFold,
}  # type: Dict[str, Type[PPRLIndex]]



class CandidateBlockingResult:
    """Object for holding candidate blocking results."""

    def __init__(self, blocks: Dict, state: PPRLIndex):
        """
        Initialise a blocking result object.
        :param blocks: A dictionary where key is set of 1 bits in bloom filter and value is a list of record IDs
        :param state: A PPRLIndex state that contains configuration of blocking
        """
        self.blocks = blocks
        self.state = state


def generate_candidate_blocks(data: Sequence[Tuple[str, ...]], signature_config: Dict, header: Optional[List[str]] = None,
                              verbose: bool = False):
    """
    :param data: list of tuples E.g. ('0', 'Kenneth Bain', '1964/06/17', 'M')
    :param signature_config:
        A description of how the signatures should be generated.
        Schema for the signature config is found in
        ``docs/schema/signature-config-schema.json``
    :param header: column names (optional)
        Program should throw exception if block features are string but header is None
    :param verbose: print additional information to std out.

    :return: A 2-tuple containing
        A list of "signatures" per record in data.
        Internal state object from the signature generation (or None).

    """
    # validate config of blocking
    validate_signature_config(signature_config)

    # extract algorithm and its config
    algorithm = signature_config.get('type', 'not specified')
    config = signature_config.get('config', 'not specified')

    # check if blocking features are column index or feature name
    blocking_features = get_config(config, 'blocking-features')
    feature_type = type(blocking_features[0])
    error_msg = 'All feature types should be the same - either feature name or feature index'
    assert all(type(x) == feature_type for x in blocking_features[1:]), error_msg

    # header should not be None if blocking features are string
    if feature_type == str:
        assert header, 'Header must not be None if blocking features are string'

    if algorithm in PPRLSTATES:
        state = PPRLSTATES[algorithm](config)
        reversed_index = state.build_reversed_index(data, verbose, header)
        state.summarize_reversed_index(reversed_index)

        # make candidate blocking result object
        candidate_block_obj = CandidateBlockingResult(reversed_index, state)

    else:
        raise NotImplementedError('The algorithm {} is not supported yet'.format(algorithm))

    return candidate_block_obj



import pkg_resources
from .pprlindex import PPRLIndex
from .pprlpsig import PPRLIndexPSignature
from .pprllambdafold import PPRLIndexLambdaFold
from .signature_generator import generate_signatures
from .blocks_generator import generate_blocks, generate_reverse_blocks
from .validation import validate_signature_config
from .candidate_blocks_generator import generate_candidate_blocks
from .encoding import generate_bloom_filter, flip_bloom_filter
from .evaluation import assess_blocks_2party

try:
    __version__ = pkg_resources.get_distribution('blocklib').version
except pkg_resources.DistributionNotFound:
    __version__ = "development"


"""Module that implement final block generations."""
from collections import defaultdict
from typing import Any, Dict, Sequence, Set, List, cast
import ast
import numpy as np

from blocklib import PPRLIndex
from .pprlpsig import PPRLIndexPSignature
from .candidate_blocks_generator import CandidateBlockingResult


def check_block_object(candidate_block_objs: Sequence[CandidateBlockingResult]):
    """
    Check candidate block objects type and their states type. Raise TypeError if conditions aren't met.
    :param candidate_block_objs: A list of candidate block result objects from 2 data providers
    :return:
    """
    for obj in candidate_block_objs:
        if type(obj) != CandidateBlockingResult:
            raise TypeError('Unsupported blocking instance {}'.format(type(obj)))

    state_type = type(candidate_block_objs[0].state)
    for obj in candidate_block_objs:
        if type(obj.state) != state_type:
            raise TypeError('Unexpected state type found: {} where we expect: {}'.format(type(obj.state), state_type))


def generate_blocks(candidate_block_objs: Sequence[CandidateBlockingResult], K: int):
    """
    Generate final blocks given list of candidate block objects from 2 or more than 2 data providers.

    :param candidate_block_objs: A list of CandidateBlockingResult from multiple data providers
    :param K: it specifies the minimum number of occurrence for records to be included in the final blocks
    :return: filtered_reversed_indices: List of dictionaries, filter out records that appear in less than K parties
    """
    check_block_object(candidate_block_objs)
    assert len(candidate_block_objs) >= K >= 2

    # obtain list of objects
    state_type = type(candidate_block_objs[0].state)
    reversed_indices = [obj.blocks for obj in candidate_block_objs]
    block_states = [obj.state for obj in candidate_block_objs]  # type: Sequence[PPRLIndex]

    filtered_reversed_indices = []  # type: List[Dict[Any, List[Any]]]
    if state_type == PPRLIndexPSignature:
        block_states = cast(Sequence[PPRLIndexPSignature], block_states)
        filtered_reversed_indices = generate_blocks_psig(reversed_indices, block_states, threshold=K)

    # default strategy: use key in reversed index as block keys
    else:
        block_keys = defaultdict(int)  # type: Dict[Any, int]
        for reversed_index in reversed_indices:
            for key in reversed_index:
                block_keys[key] += 1
        final_block_keys = [key for key, count in block_keys.items() if count >= K]
        for reversed_index in reversed_indices:
            reversed_index = {k: v for k, v in reversed_index.items() if k in final_block_keys}
            filtered_reversed_indices.append(reversed_index)

    return filtered_reversed_indices


def generate_reverse_blocks(reversed_indices: Sequence[Dict]):
    """
    Return a list of dictionaries of record to block key mapping
    :param reversed_indices: A list of dictionaries where key is the block key and value is a list of record IDs.
    :return: rec_to_blockkey: A list of dictionaries where key is the record ID and value is a set of block key the record belongs to
    """
    rec_to_blockkey = []
    for reversed_index in reversed_indices:
        map_rec_block = defaultdict(set)  # type: Dict[Any, Set[Any]]
        for blk_key, rec_list in reversed_index.items():
            for rec in rec_list:
                map_rec_block[rec].add(blk_key)
        rec_to_blockkey.append(map_rec_block)
    return rec_to_blockkey


def generate_blocks_psig(reversed_indices: Sequence[Dict], block_states: Sequence[PPRLIndexPSignature], threshold: int):
    """
    Generate final blocks for P-Sig.
    :param reversed_indices: A list of dictionaries where key is the block key and value is a list of record IDs.
    :param block_states: A list of PPRLIndex objects that hold configuration of the blocking job
    :param threshold: int which decides a pair when number of 1 bits in bloom filter is large than or equal to threshold
    :return: reversed_indices: A list of dictionaries where blocks that don't contain any matches are deleted
    """
    # generate candidate bloom filters
    candidate_bloom_filters = []
    for reversed_index, state in zip(reversed_indices, block_states):
        cbf = set()  # type: Set[int]
        for bf_set in reversed_index:
            bf_set = ast.literal_eval(bf_set)
            cbf = cbf.union(bf_set)

        bf_len = int(block_states[0].blocking_config.get("bf-len", None))
        bf_vector = np.zeros(bf_len, dtype=bool)
        bf_vector[list(cbf)] = True
        candidate_bloom_filters.append(bf_vector)

    # compute blocking filter (and operation)
    cbf_array = np.sum(candidate_bloom_filters, axis=0)
    block_filter = cbf_array >= threshold

    # filter reversed_indices with block filter
    for reversed_index in reversed_indices:
        has_matches = {bf_set: all(block_filter[i] for i in ast.literal_eval(bf_set)) for bf_set in reversed_index}
        for bf_set in has_matches:
            if not has_matches[bf_set]:
                del reversed_index[bf_set]

    # because of collisions in counting bloom filter, there are blocks only unique to one filtered index
    # only keep blocks that exist in at least threshold many reversed indices
    keys = defaultdict(int)  # type: Dict[Set, int]
    for reversed_index in reversed_indices:
        for k in reversed_index:
            keys[k] += 1
    common_keys = [k for k in keys if keys[k] >= threshold]
    clean_reversed_indices = []  # type: List[Dict[Set, List]]
    for reversed_index in reversed_indices:
        clean_reversed_indices.append(dict((k, reversed_index[k]) for k in common_keys if k in reversed_index))

    return clean_reversed_indices
